# Author: xuejinlin
# Date: 2023/3/7 16:16

categories = [
        {
            "id": 0,
            "name": "normal",
            "reward": 0.0
        },{
            "id": 1,
            "name": "status_zhugong",
            "reward": 0.01
        }, {
            "id": 3,
            "name": "status_jishaxiaobing",
            "reward": 0.05
        }, {
            "id": 4,
            "name": "death",
            "reward": -0.02
        }, {
            "id": 6,
            "name": "status_beitagongji",
            "reward": -0.01
        }, {
            "id": 10,
            "name": "status_jishayx",
            "reward": 0.03
        }, {
            "id": 13,
            "name": "victory",
            "reward": 0.06
        },{
            "id": 15,
            "name": "defeated",
            "reward": -0.06
        }
    ]

